/* Do not edit this file. It is produced from the corresponding .m4 source */
/*
 *  Copyright (C) 2014, Northwestern University and Argonne National Laboratory
 *  See COPYRIGHT notice in top-level directory.
 */
/* $Id: convert_swap.m4 2290 2016-01-02 18:37:46Z wkliao $ */

#if HAVE_CONFIG_H
# include "ncconfig.h"
#endif

#include <stdio.h>
#include <unistd.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#include <assert.h>
#include <arpa/inet.h>   /* htonl(), htons() */

#include <mpi.h>

#include "nc.h"
#include "ncx.h"
#include "macro.h"

/* Prototypes for functions used only in this file */
#if 0
static void swapn(void *dst, const void *src, MPI_Offset nn, int xsize);
#endif

/*
 *  Datatype Mapping:
 *
 *  NETCDF    <--> MPI                    Description
 *   NC_BYTE       MPI_SIGNED_CHAR        signed 1-byte integer
 *   NC_CHAR       MPI_CHAR               char, text (cannot convert to other types)
 *   NC_SHORT      MPI_SHORT              signed 2-byte integer
 *   NC_INT        MPI_INT                signed 4-byte integer
 *   NC_FLOAT      MPI_FLOAT              single precision floating point
 *   NC_DOUBLE     MPI_DOUBLE             double precision floating point
 *   NC_UBYTE      MPI_UNSIGNED_CHAR      unsigned 1-byte int
 *   NC_USHORT     MPI_UNSIGNED_SHORT     unsigned 2-byte int
 *   NC_UINT       MPI_UNSIGNED           unsigned 4-byte int
 *   NC_INT64      MPI_LONG_LONG_INT      signed 8-byte int
 *   NC_UINT64     MPI_UNSIGNED_LONG_LONG unsigned 8-byte int
 *
 *  Assume: MPI_Datatype and nc_type are both enumerable types
 *          (this might not conform with MPI, as MPI_Datatype is intended to be
 *           an opaque data type.)
 *
 *  In OpenMPI, this assumption will fail
 */

inline MPI_Datatype
ncmpii_nc2mpitype(nc_type type)
{
    switch(type){
        case NC_BYTE :   return MPI_SIGNED_CHAR;
        case NC_CHAR :   return MPI_CHAR;
        case NC_SHORT :  return MPI_SHORT;
        case NC_INT :    return MPI_INT;
        case NC_FLOAT :  return MPI_FLOAT;
        case NC_DOUBLE : return MPI_DOUBLE;
        case NC_UBYTE :  return MPI_UNSIGNED_CHAR;
        case NC_USHORT : return MPI_UNSIGNED_SHORT;
        case NC_UINT :   return MPI_UNSIGNED;
        case NC_INT64 :  return MPI_LONG_LONG_INT;
        case NC_UINT64 : return MPI_UNSIGNED_LONG_LONG;
        default:         return MPI_DATATYPE_NULL;
    }
}

/*----< ncmpii_need_convert() >----------------------------------------------*/
/* netCDF specification make a special case for type conversion between
 * uchar and scahr: do not check for range error. See
 * http://www.unidata.ucar.edu/software/netcdf/docs_rc/data_type.html#type_conversion
 */
inline int
ncmpii_need_convert(nc_type nctype,MPI_Datatype mpitype) {
    return !( (nctype == NC_CHAR   && mpitype == MPI_CHAR)           ||
              (nctype == NC_BYTE   && mpitype == MPI_SIGNED_CHAR)    ||
              (nctype == NC_BYTE   && mpitype == MPI_UNSIGNED_CHAR)  ||
#if defined(__CHAR_UNSIGNED__) && __CHAR_UNSIGNED__ != 0
              (nctype == NC_BYTE   && mpitype == MPI_CHAR)           ||
#endif
              (nctype == NC_SHORT  && mpitype == MPI_SHORT)          ||
              (nctype == NC_INT    && mpitype == MPI_INT)            ||
              (nctype == NC_INT    && mpitype == MPI_LONG &&
               X_SIZEOF_INT == SIZEOF_LONG)                          ||
              (nctype == NC_FLOAT  && mpitype == MPI_FLOAT)          ||
              (nctype == NC_DOUBLE && mpitype == MPI_DOUBLE)         ||
              (nctype == NC_UBYTE  && mpitype == MPI_UNSIGNED_CHAR)  ||
              (nctype == NC_USHORT && mpitype == MPI_UNSIGNED_SHORT) ||
              (nctype == NC_UINT   && mpitype == MPI_UNSIGNED)       ||
              (nctype == NC_INT64  && mpitype == MPI_LONG_LONG_INT)  ||
              (nctype == NC_UINT64 && mpitype == MPI_UNSIGNED_LONG_LONG)
            );
}

/*----< ncmpii_need_swap() >-------------------------------------------------*/
inline int
ncmpii_need_swap(nc_type      nctype,
                 MPI_Datatype mpitype)
{
#ifdef WORDS_BIGENDIAN
    return 0;
#else
    if ((nctype == NC_CHAR   && mpitype == MPI_CHAR)           ||
        (nctype == NC_BYTE   && mpitype == MPI_SIGNED_CHAR)    ||
        (nctype == NC_UBYTE  && mpitype == MPI_UNSIGNED_CHAR))
        return 0;

    return 1;
#endif
}

#if 0
/*----< swapn() >------------------------------------------------------------*/
inline static void
swapn(void       *dst,
      const void *src,
      MPI_Offset  nn,
      int         xsize)
{
    int i;
    uchar *op = dst;
    const uchar *ip = src;
    while (nn-- != 0) {
        for (i=0; i<xsize; i++)
            op[i] = ip[xsize-1-i];
        op += xsize;
        ip += xsize;
    }
}
#endif

/* Endianness byte swap: done in-place */
#define SWAP(x,y) {tmp = (x); (x) = (y); (y) = tmp;}

/*----< ncmpii_swap() >-------------------------------------------------------*/
void
ncmpii_swapn(void       *dest_p,  /* destination array */
             const void *src_p,   /* source array */
             MPI_Offset  nelems,  /* number of elements in buf[] */
             int         esize)   /* byte size of each element */
{
    int  i;

    if (esize <= 1 || nelems <= 0) return;  /* no need */

    if (esize == 4) { /* this is the most common case */
              uint32_t *dest = (uint32_t*)       dest_p;
        const uint32_t *src  = (const uint32_t*) src_p;
        for (i=0; i<nelems; i++)
            dest[i] = htonl(src[i]);
    }
    else if (esize == 2) {
              uint16_t *dest =       (uint16_t*) dest_p;
        const uint16_t *src  = (const uint16_t*) src_p;
        for (i=0; i<nelems; i++)
            dest[i] = htons(src[i]);
    }
    else {
              uchar *op = (uchar*) dest_p;
        const uchar *ip = (uchar*) src_p;
        /* for esize is not 1, 2, or 4 */
        while (nelems-- > 0) {
            for (i=0; i<esize; i++)
                op[i] = ip[esize-1-i];
            op += esize;
            ip += esize;
        }
    }
}

/*----< ncmpii_in_swap() >---------------------------------------------------*/
void
ncmpii_in_swapn(void       *buf,
                MPI_Offset  nelems,  /* number of elements in buf[] */
                int         esize)   /* byte size of each element */
{
    int  i;
    uchar tmp, *op = (uchar*)buf;

    if (esize <= 1 || nelems <= 0) return;  /* no need */

    if (esize == 4) { /* this is the most common case */
        uint32_t *dest = (uint32_t*) buf;
        for (i=0; i<nelems; i++)
            dest[i] = htonl(dest[i]);
    }
    else if (esize == 2) {
        uint16_t *dest = (uint16_t*) buf;
        for (i=0; i<nelems; i++)
            dest[i] = htons(dest[i]);
    }
    else {
        /* for esize is not 1, 2, or 4 */
        while (nelems-- > 0) {
            for (i=0; i<esize/2; i++)
                SWAP(op[i], op[esize-1-i])
            op += esize;
        }
    }
}



/*----< ncmpii_x_putn_schar() >--------------------------------------------------*/
inline int
ncmpii_x_putn_schar(void         *xp,      /* file buffer of type schar */
                 const void   *putbuf,  /* put buffer of type puttype */
                 MPI_Offset    nelems,
                 MPI_Datatype  puttype)
{
    if (puttype == MPI_CHAR || /* assume ECHAR has been checked before */
        puttype == MPI_SIGNED_CHAR)
        return ncmpix_putn_schar_schar    (&xp, nelems, (const schar*)     putbuf);
    else if (puttype == MPI_UNSIGNED_CHAR)
        return ncmpix_putn_schar_uchar    (&xp, nelems, (const uchar*)     putbuf);
    else if (puttype == MPI_SHORT)
        return ncmpix_putn_schar_short    (&xp, nelems, (const short*)     putbuf);
    else if (puttype == MPI_UNSIGNED_SHORT)
        return ncmpix_putn_schar_ushort   (&xp, nelems, (const ushort*)    putbuf);
    else if (puttype == MPI_INT)
        return ncmpix_putn_schar_int      (&xp, nelems, (const int*)       putbuf);
    else if (puttype == MPI_UNSIGNED)
        return ncmpix_putn_schar_uint     (&xp, nelems, (const uint*)      putbuf);
    else if (puttype == MPI_LONG)
        return ncmpix_putn_schar_long     (&xp, nelems, (const long*)      putbuf);
    else if (puttype == MPI_FLOAT)
        return ncmpix_putn_schar_float    (&xp, nelems, (const float*)     putbuf);
    else if (puttype == MPI_DOUBLE)
        return ncmpix_putn_schar_double   (&xp, nelems, (const double*)    putbuf);
    else if (puttype == MPI_LONG_LONG_INT)
        return ncmpix_putn_schar_longlong (&xp, nelems, (const longlong*)  putbuf);
    else if (puttype == MPI_UNSIGNED_LONG_LONG)
        return ncmpix_putn_schar_ulonglong(&xp, nelems, (const ulonglong*) putbuf);
    DEBUG_RETURN_ERROR(NC_EBADTYPE)
}

/*----< ncmpii_x_putn_uchar() >--------------------------------------------------*/
inline int
ncmpii_x_putn_uchar(void         *xp,      /* file buffer of type schar */
                 const void   *putbuf,  /* put buffer of type puttype */
                 MPI_Offset    nelems,
                 MPI_Datatype  puttype)
{
    if (puttype == MPI_CHAR || /* assume ECHAR has been checked before */
        puttype == MPI_SIGNED_CHAR)
        return ncmpix_putn_uchar_schar    (&xp, nelems, (const schar*)     putbuf);
    else if (puttype == MPI_UNSIGNED_CHAR)
        return ncmpix_putn_uchar_uchar    (&xp, nelems, (const uchar*)     putbuf);
    else if (puttype == MPI_SHORT)
        return ncmpix_putn_uchar_short    (&xp, nelems, (const short*)     putbuf);
    else if (puttype == MPI_UNSIGNED_SHORT)
        return ncmpix_putn_uchar_ushort   (&xp, nelems, (const ushort*)    putbuf);
    else if (puttype == MPI_INT)
        return ncmpix_putn_uchar_int      (&xp, nelems, (const int*)       putbuf);
    else if (puttype == MPI_UNSIGNED)
        return ncmpix_putn_uchar_uint     (&xp, nelems, (const uint*)      putbuf);
    else if (puttype == MPI_LONG)
        return ncmpix_putn_uchar_long     (&xp, nelems, (const long*)      putbuf);
    else if (puttype == MPI_FLOAT)
        return ncmpix_putn_uchar_float    (&xp, nelems, (const float*)     putbuf);
    else if (puttype == MPI_DOUBLE)
        return ncmpix_putn_uchar_double   (&xp, nelems, (const double*)    putbuf);
    else if (puttype == MPI_LONG_LONG_INT)
        return ncmpix_putn_uchar_longlong (&xp, nelems, (const longlong*)  putbuf);
    else if (puttype == MPI_UNSIGNED_LONG_LONG)
        return ncmpix_putn_uchar_ulonglong(&xp, nelems, (const ulonglong*) putbuf);
    DEBUG_RETURN_ERROR(NC_EBADTYPE)
}

/*----< ncmpii_x_putn_short() >--------------------------------------------------*/
inline int
ncmpii_x_putn_short(void         *xp,      /* file buffer of type schar */
                 const void   *putbuf,  /* put buffer of type puttype */
                 MPI_Offset    nelems,
                 MPI_Datatype  puttype)
{
    if (puttype == MPI_CHAR || /* assume ECHAR has been checked before */
        puttype == MPI_SIGNED_CHAR)
        return ncmpix_putn_short_schar    (&xp, nelems, (const schar*)     putbuf);
    else if (puttype == MPI_UNSIGNED_CHAR)
        return ncmpix_putn_short_uchar    (&xp, nelems, (const uchar*)     putbuf);
    else if (puttype == MPI_SHORT)
        return ncmpix_putn_short_short    (&xp, nelems, (const short*)     putbuf);
    else if (puttype == MPI_UNSIGNED_SHORT)
        return ncmpix_putn_short_ushort   (&xp, nelems, (const ushort*)    putbuf);
    else if (puttype == MPI_INT)
        return ncmpix_putn_short_int      (&xp, nelems, (const int*)       putbuf);
    else if (puttype == MPI_UNSIGNED)
        return ncmpix_putn_short_uint     (&xp, nelems, (const uint*)      putbuf);
    else if (puttype == MPI_LONG)
        return ncmpix_putn_short_long     (&xp, nelems, (const long*)      putbuf);
    else if (puttype == MPI_FLOAT)
        return ncmpix_putn_short_float    (&xp, nelems, (const float*)     putbuf);
    else if (puttype == MPI_DOUBLE)
        return ncmpix_putn_short_double   (&xp, nelems, (const double*)    putbuf);
    else if (puttype == MPI_LONG_LONG_INT)
        return ncmpix_putn_short_longlong (&xp, nelems, (const longlong*)  putbuf);
    else if (puttype == MPI_UNSIGNED_LONG_LONG)
        return ncmpix_putn_short_ulonglong(&xp, nelems, (const ulonglong*) putbuf);
    DEBUG_RETURN_ERROR(NC_EBADTYPE)
}

/*----< ncmpii_x_putn_ushort() >--------------------------------------------------*/
inline int
ncmpii_x_putn_ushort(void         *xp,      /* file buffer of type schar */
                 const void   *putbuf,  /* put buffer of type puttype */
                 MPI_Offset    nelems,
                 MPI_Datatype  puttype)
{
    if (puttype == MPI_CHAR || /* assume ECHAR has been checked before */
        puttype == MPI_SIGNED_CHAR)
        return ncmpix_putn_ushort_schar    (&xp, nelems, (const schar*)     putbuf);
    else if (puttype == MPI_UNSIGNED_CHAR)
        return ncmpix_putn_ushort_uchar    (&xp, nelems, (const uchar*)     putbuf);
    else if (puttype == MPI_SHORT)
        return ncmpix_putn_ushort_short    (&xp, nelems, (const short*)     putbuf);
    else if (puttype == MPI_UNSIGNED_SHORT)
        return ncmpix_putn_ushort_ushort   (&xp, nelems, (const ushort*)    putbuf);
    else if (puttype == MPI_INT)
        return ncmpix_putn_ushort_int      (&xp, nelems, (const int*)       putbuf);
    else if (puttype == MPI_UNSIGNED)
        return ncmpix_putn_ushort_uint     (&xp, nelems, (const uint*)      putbuf);
    else if (puttype == MPI_LONG)
        return ncmpix_putn_ushort_long     (&xp, nelems, (const long*)      putbuf);
    else if (puttype == MPI_FLOAT)
        return ncmpix_putn_ushort_float    (&xp, nelems, (const float*)     putbuf);
    else if (puttype == MPI_DOUBLE)
        return ncmpix_putn_ushort_double   (&xp, nelems, (const double*)    putbuf);
    else if (puttype == MPI_LONG_LONG_INT)
        return ncmpix_putn_ushort_longlong (&xp, nelems, (const longlong*)  putbuf);
    else if (puttype == MPI_UNSIGNED_LONG_LONG)
        return ncmpix_putn_ushort_ulonglong(&xp, nelems, (const ulonglong*) putbuf);
    DEBUG_RETURN_ERROR(NC_EBADTYPE)
}

/*----< ncmpii_x_putn_int() >--------------------------------------------------*/
inline int
ncmpii_x_putn_int(void         *xp,      /* file buffer of type schar */
                 const void   *putbuf,  /* put buffer of type puttype */
                 MPI_Offset    nelems,
                 MPI_Datatype  puttype)
{
    if (puttype == MPI_CHAR || /* assume ECHAR has been checked before */
        puttype == MPI_SIGNED_CHAR)
        return ncmpix_putn_int_schar    (&xp, nelems, (const schar*)     putbuf);
    else if (puttype == MPI_UNSIGNED_CHAR)
        return ncmpix_putn_int_uchar    (&xp, nelems, (const uchar*)     putbuf);
    else if (puttype == MPI_SHORT)
        return ncmpix_putn_int_short    (&xp, nelems, (const short*)     putbuf);
    else if (puttype == MPI_UNSIGNED_SHORT)
        return ncmpix_putn_int_ushort   (&xp, nelems, (const ushort*)    putbuf);
    else if (puttype == MPI_INT)
        return ncmpix_putn_int_int      (&xp, nelems, (const int*)       putbuf);
    else if (puttype == MPI_UNSIGNED)
        return ncmpix_putn_int_uint     (&xp, nelems, (const uint*)      putbuf);
    else if (puttype == MPI_LONG)
        return ncmpix_putn_int_long     (&xp, nelems, (const long*)      putbuf);
    else if (puttype == MPI_FLOAT)
        return ncmpix_putn_int_float    (&xp, nelems, (const float*)     putbuf);
    else if (puttype == MPI_DOUBLE)
        return ncmpix_putn_int_double   (&xp, nelems, (const double*)    putbuf);
    else if (puttype == MPI_LONG_LONG_INT)
        return ncmpix_putn_int_longlong (&xp, nelems, (const longlong*)  putbuf);
    else if (puttype == MPI_UNSIGNED_LONG_LONG)
        return ncmpix_putn_int_ulonglong(&xp, nelems, (const ulonglong*) putbuf);
    DEBUG_RETURN_ERROR(NC_EBADTYPE)
}

/*----< ncmpii_x_putn_uint() >--------------------------------------------------*/
inline int
ncmpii_x_putn_uint(void         *xp,      /* file buffer of type schar */
                 const void   *putbuf,  /* put buffer of type puttype */
                 MPI_Offset    nelems,
                 MPI_Datatype  puttype)
{
    if (puttype == MPI_CHAR || /* assume ECHAR has been checked before */
        puttype == MPI_SIGNED_CHAR)
        return ncmpix_putn_uint_schar    (&xp, nelems, (const schar*)     putbuf);
    else if (puttype == MPI_UNSIGNED_CHAR)
        return ncmpix_putn_uint_uchar    (&xp, nelems, (const uchar*)     putbuf);
    else if (puttype == MPI_SHORT)
        return ncmpix_putn_uint_short    (&xp, nelems, (const short*)     putbuf);
    else if (puttype == MPI_UNSIGNED_SHORT)
        return ncmpix_putn_uint_ushort   (&xp, nelems, (const ushort*)    putbuf);
    else if (puttype == MPI_INT)
        return ncmpix_putn_uint_int      (&xp, nelems, (const int*)       putbuf);
    else if (puttype == MPI_UNSIGNED)
        return ncmpix_putn_uint_uint     (&xp, nelems, (const uint*)      putbuf);
    else if (puttype == MPI_LONG)
        return ncmpix_putn_uint_long     (&xp, nelems, (const long*)      putbuf);
    else if (puttype == MPI_FLOAT)
        return ncmpix_putn_uint_float    (&xp, nelems, (const float*)     putbuf);
    else if (puttype == MPI_DOUBLE)
        return ncmpix_putn_uint_double   (&xp, nelems, (const double*)    putbuf);
    else if (puttype == MPI_LONG_LONG_INT)
        return ncmpix_putn_uint_longlong (&xp, nelems, (const longlong*)  putbuf);
    else if (puttype == MPI_UNSIGNED_LONG_LONG)
        return ncmpix_putn_uint_ulonglong(&xp, nelems, (const ulonglong*) putbuf);
    DEBUG_RETURN_ERROR(NC_EBADTYPE)
}

/*----< ncmpii_x_putn_float() >--------------------------------------------------*/
inline int
ncmpii_x_putn_float(void         *xp,      /* file buffer of type schar */
                 const void   *putbuf,  /* put buffer of type puttype */
                 MPI_Offset    nelems,
                 MPI_Datatype  puttype)
{
    if (puttype == MPI_CHAR || /* assume ECHAR has been checked before */
        puttype == MPI_SIGNED_CHAR)
        return ncmpix_putn_float_schar    (&xp, nelems, (const schar*)     putbuf);
    else if (puttype == MPI_UNSIGNED_CHAR)
        return ncmpix_putn_float_uchar    (&xp, nelems, (const uchar*)     putbuf);
    else if (puttype == MPI_SHORT)
        return ncmpix_putn_float_short    (&xp, nelems, (const short*)     putbuf);
    else if (puttype == MPI_UNSIGNED_SHORT)
        return ncmpix_putn_float_ushort   (&xp, nelems, (const ushort*)    putbuf);
    else if (puttype == MPI_INT)
        return ncmpix_putn_float_int      (&xp, nelems, (const int*)       putbuf);
    else if (puttype == MPI_UNSIGNED)
        return ncmpix_putn_float_uint     (&xp, nelems, (const uint*)      putbuf);
    else if (puttype == MPI_LONG)
        return ncmpix_putn_float_long     (&xp, nelems, (const long*)      putbuf);
    else if (puttype == MPI_FLOAT)
        return ncmpix_putn_float_float    (&xp, nelems, (const float*)     putbuf);
    else if (puttype == MPI_DOUBLE)
        return ncmpix_putn_float_double   (&xp, nelems, (const double*)    putbuf);
    else if (puttype == MPI_LONG_LONG_INT)
        return ncmpix_putn_float_longlong (&xp, nelems, (const longlong*)  putbuf);
    else if (puttype == MPI_UNSIGNED_LONG_LONG)
        return ncmpix_putn_float_ulonglong(&xp, nelems, (const ulonglong*) putbuf);
    DEBUG_RETURN_ERROR(NC_EBADTYPE)
}

/*----< ncmpii_x_putn_double() >--------------------------------------------------*/
inline int
ncmpii_x_putn_double(void         *xp,      /* file buffer of type schar */
                 const void   *putbuf,  /* put buffer of type puttype */
                 MPI_Offset    nelems,
                 MPI_Datatype  puttype)
{
    if (puttype == MPI_CHAR || /* assume ECHAR has been checked before */
        puttype == MPI_SIGNED_CHAR)
        return ncmpix_putn_double_schar    (&xp, nelems, (const schar*)     putbuf);
    else if (puttype == MPI_UNSIGNED_CHAR)
        return ncmpix_putn_double_uchar    (&xp, nelems, (const uchar*)     putbuf);
    else if (puttype == MPI_SHORT)
        return ncmpix_putn_double_short    (&xp, nelems, (const short*)     putbuf);
    else if (puttype == MPI_UNSIGNED_SHORT)
        return ncmpix_putn_double_ushort   (&xp, nelems, (const ushort*)    putbuf);
    else if (puttype == MPI_INT)
        return ncmpix_putn_double_int      (&xp, nelems, (const int*)       putbuf);
    else if (puttype == MPI_UNSIGNED)
        return ncmpix_putn_double_uint     (&xp, nelems, (const uint*)      putbuf);
    else if (puttype == MPI_LONG)
        return ncmpix_putn_double_long     (&xp, nelems, (const long*)      putbuf);
    else if (puttype == MPI_FLOAT)
        return ncmpix_putn_double_float    (&xp, nelems, (const float*)     putbuf);
    else if (puttype == MPI_DOUBLE)
        return ncmpix_putn_double_double   (&xp, nelems, (const double*)    putbuf);
    else if (puttype == MPI_LONG_LONG_INT)
        return ncmpix_putn_double_longlong (&xp, nelems, (const longlong*)  putbuf);
    else if (puttype == MPI_UNSIGNED_LONG_LONG)
        return ncmpix_putn_double_ulonglong(&xp, nelems, (const ulonglong*) putbuf);
    DEBUG_RETURN_ERROR(NC_EBADTYPE)
}

/*----< ncmpii_x_putn_int64() >--------------------------------------------------*/
inline int
ncmpii_x_putn_int64(void         *xp,      /* file buffer of type schar */
                 const void   *putbuf,  /* put buffer of type puttype */
                 MPI_Offset    nelems,
                 MPI_Datatype  puttype)
{
    if (puttype == MPI_CHAR || /* assume ECHAR has been checked before */
        puttype == MPI_SIGNED_CHAR)
        return ncmpix_putn_int64_schar    (&xp, nelems, (const schar*)     putbuf);
    else if (puttype == MPI_UNSIGNED_CHAR)
        return ncmpix_putn_int64_uchar    (&xp, nelems, (const uchar*)     putbuf);
    else if (puttype == MPI_SHORT)
        return ncmpix_putn_int64_short    (&xp, nelems, (const short*)     putbuf);
    else if (puttype == MPI_UNSIGNED_SHORT)
        return ncmpix_putn_int64_ushort   (&xp, nelems, (const ushort*)    putbuf);
    else if (puttype == MPI_INT)
        return ncmpix_putn_int64_int      (&xp, nelems, (const int*)       putbuf);
    else if (puttype == MPI_UNSIGNED)
        return ncmpix_putn_int64_uint     (&xp, nelems, (const uint*)      putbuf);
    else if (puttype == MPI_LONG)
        return ncmpix_putn_int64_long     (&xp, nelems, (const long*)      putbuf);
    else if (puttype == MPI_FLOAT)
        return ncmpix_putn_int64_float    (&xp, nelems, (const float*)     putbuf);
    else if (puttype == MPI_DOUBLE)
        return ncmpix_putn_int64_double   (&xp, nelems, (const double*)    putbuf);
    else if (puttype == MPI_LONG_LONG_INT)
        return ncmpix_putn_int64_longlong (&xp, nelems, (const longlong*)  putbuf);
    else if (puttype == MPI_UNSIGNED_LONG_LONG)
        return ncmpix_putn_int64_ulonglong(&xp, nelems, (const ulonglong*) putbuf);
    DEBUG_RETURN_ERROR(NC_EBADTYPE)
}

/*----< ncmpii_x_putn_uint64() >--------------------------------------------------*/
inline int
ncmpii_x_putn_uint64(void         *xp,      /* file buffer of type schar */
                 const void   *putbuf,  /* put buffer of type puttype */
                 MPI_Offset    nelems,
                 MPI_Datatype  puttype)
{
    if (puttype == MPI_CHAR || /* assume ECHAR has been checked before */
        puttype == MPI_SIGNED_CHAR)
        return ncmpix_putn_uint64_schar    (&xp, nelems, (const schar*)     putbuf);
    else if (puttype == MPI_UNSIGNED_CHAR)
        return ncmpix_putn_uint64_uchar    (&xp, nelems, (const uchar*)     putbuf);
    else if (puttype == MPI_SHORT)
        return ncmpix_putn_uint64_short    (&xp, nelems, (const short*)     putbuf);
    else if (puttype == MPI_UNSIGNED_SHORT)
        return ncmpix_putn_uint64_ushort   (&xp, nelems, (const ushort*)    putbuf);
    else if (puttype == MPI_INT)
        return ncmpix_putn_uint64_int      (&xp, nelems, (const int*)       putbuf);
    else if (puttype == MPI_UNSIGNED)
        return ncmpix_putn_uint64_uint     (&xp, nelems, (const uint*)      putbuf);
    else if (puttype == MPI_LONG)
        return ncmpix_putn_uint64_long     (&xp, nelems, (const long*)      putbuf);
    else if (puttype == MPI_FLOAT)
        return ncmpix_putn_uint64_float    (&xp, nelems, (const float*)     putbuf);
    else if (puttype == MPI_DOUBLE)
        return ncmpix_putn_uint64_double   (&xp, nelems, (const double*)    putbuf);
    else if (puttype == MPI_LONG_LONG_INT)
        return ncmpix_putn_uint64_longlong (&xp, nelems, (const longlong*)  putbuf);
    else if (puttype == MPI_UNSIGNED_LONG_LONG)
        return ncmpix_putn_uint64_ulonglong(&xp, nelems, (const ulonglong*) putbuf);
    DEBUG_RETURN_ERROR(NC_EBADTYPE)
}



/*----< ncmpii_x_getn_schar() >-------------------------------------------------*/
inline int
ncmpii_x_getn_schar(const void   *xp,      /* file buffer of type schar */
                 void         *getbuf,  /* get buffer of type gettype */
                 MPI_Offset    nelems,
                 MPI_Datatype  gettype)
{
    if (gettype == MPI_CHAR || /* assume ECHAR has been checked before */
        gettype == MPI_SIGNED_CHAR)
        return ncmpix_getn_schar_schar    (&xp, nelems, (schar*)     getbuf);
    else if (gettype == MPI_UNSIGNED_CHAR)
        return ncmpix_getn_schar_uchar    (&xp, nelems, (uchar*)     getbuf);
    else if (gettype == MPI_SHORT)
        return ncmpix_getn_schar_short    (&xp, nelems, (short*)     getbuf);
    else if (gettype == MPI_UNSIGNED_SHORT)
        return ncmpix_getn_schar_ushort   (&xp, nelems, (ushort*)    getbuf);
    else if (gettype == MPI_INT)
        return ncmpix_getn_schar_int      (&xp, nelems, (int*)       getbuf);
    else if (gettype == MPI_UNSIGNED)
        return ncmpix_getn_schar_uint     (&xp, nelems, (uint*)      getbuf);
    else if (gettype == MPI_LONG)
        return ncmpix_getn_schar_long     (&xp, nelems, (long*)      getbuf);
    else if (gettype == MPI_FLOAT)
        return ncmpix_getn_schar_float    (&xp, nelems, (float*)     getbuf);
    else if (gettype == MPI_DOUBLE)
        return ncmpix_getn_schar_double   (&xp, nelems, (double*)    getbuf);
    else if (gettype == MPI_LONG_LONG_INT)
        return ncmpix_getn_schar_longlong (&xp, nelems, (longlong*)  getbuf);
    else if (gettype == MPI_UNSIGNED_LONG_LONG)
        return ncmpix_getn_schar_ulonglong(&xp, nelems, (ulonglong*) getbuf);
    DEBUG_RETURN_ERROR(NC_EBADTYPE)
}

/*----< ncmpii_x_getn_uchar() >-------------------------------------------------*/
inline int
ncmpii_x_getn_uchar(const void   *xp,      /* file buffer of type schar */
                 void         *getbuf,  /* get buffer of type gettype */
                 MPI_Offset    nelems,
                 MPI_Datatype  gettype)
{
    if (gettype == MPI_CHAR || /* assume ECHAR has been checked before */
        gettype == MPI_SIGNED_CHAR)
        return ncmpix_getn_uchar_schar    (&xp, nelems, (schar*)     getbuf);
    else if (gettype == MPI_UNSIGNED_CHAR)
        return ncmpix_getn_uchar_uchar    (&xp, nelems, (uchar*)     getbuf);
    else if (gettype == MPI_SHORT)
        return ncmpix_getn_uchar_short    (&xp, nelems, (short*)     getbuf);
    else if (gettype == MPI_UNSIGNED_SHORT)
        return ncmpix_getn_uchar_ushort   (&xp, nelems, (ushort*)    getbuf);
    else if (gettype == MPI_INT)
        return ncmpix_getn_uchar_int      (&xp, nelems, (int*)       getbuf);
    else if (gettype == MPI_UNSIGNED)
        return ncmpix_getn_uchar_uint     (&xp, nelems, (uint*)      getbuf);
    else if (gettype == MPI_LONG)
        return ncmpix_getn_uchar_long     (&xp, nelems, (long*)      getbuf);
    else if (gettype == MPI_FLOAT)
        return ncmpix_getn_uchar_float    (&xp, nelems, (float*)     getbuf);
    else if (gettype == MPI_DOUBLE)
        return ncmpix_getn_uchar_double   (&xp, nelems, (double*)    getbuf);
    else if (gettype == MPI_LONG_LONG_INT)
        return ncmpix_getn_uchar_longlong (&xp, nelems, (longlong*)  getbuf);
    else if (gettype == MPI_UNSIGNED_LONG_LONG)
        return ncmpix_getn_uchar_ulonglong(&xp, nelems, (ulonglong*) getbuf);
    DEBUG_RETURN_ERROR(NC_EBADTYPE)
}

/*----< ncmpii_x_getn_short() >-------------------------------------------------*/
inline int
ncmpii_x_getn_short(const void   *xp,      /* file buffer of type schar */
                 void         *getbuf,  /* get buffer of type gettype */
                 MPI_Offset    nelems,
                 MPI_Datatype  gettype)
{
    if (gettype == MPI_CHAR || /* assume ECHAR has been checked before */
        gettype == MPI_SIGNED_CHAR)
        return ncmpix_getn_short_schar    (&xp, nelems, (schar*)     getbuf);
    else if (gettype == MPI_UNSIGNED_CHAR)
        return ncmpix_getn_short_uchar    (&xp, nelems, (uchar*)     getbuf);
    else if (gettype == MPI_SHORT)
        return ncmpix_getn_short_short    (&xp, nelems, (short*)     getbuf);
    else if (gettype == MPI_UNSIGNED_SHORT)
        return ncmpix_getn_short_ushort   (&xp, nelems, (ushort*)    getbuf);
    else if (gettype == MPI_INT)
        return ncmpix_getn_short_int      (&xp, nelems, (int*)       getbuf);
    else if (gettype == MPI_UNSIGNED)
        return ncmpix_getn_short_uint     (&xp, nelems, (uint*)      getbuf);
    else if (gettype == MPI_LONG)
        return ncmpix_getn_short_long     (&xp, nelems, (long*)      getbuf);
    else if (gettype == MPI_FLOAT)
        return ncmpix_getn_short_float    (&xp, nelems, (float*)     getbuf);
    else if (gettype == MPI_DOUBLE)
        return ncmpix_getn_short_double   (&xp, nelems, (double*)    getbuf);
    else if (gettype == MPI_LONG_LONG_INT)
        return ncmpix_getn_short_longlong (&xp, nelems, (longlong*)  getbuf);
    else if (gettype == MPI_UNSIGNED_LONG_LONG)
        return ncmpix_getn_short_ulonglong(&xp, nelems, (ulonglong*) getbuf);
    DEBUG_RETURN_ERROR(NC_EBADTYPE)
}

/*----< ncmpii_x_getn_ushort() >-------------------------------------------------*/
inline int
ncmpii_x_getn_ushort(const void   *xp,      /* file buffer of type schar */
                 void         *getbuf,  /* get buffer of type gettype */
                 MPI_Offset    nelems,
                 MPI_Datatype  gettype)
{
    if (gettype == MPI_CHAR || /* assume ECHAR has been checked before */
        gettype == MPI_SIGNED_CHAR)
        return ncmpix_getn_ushort_schar    (&xp, nelems, (schar*)     getbuf);
    else if (gettype == MPI_UNSIGNED_CHAR)
        return ncmpix_getn_ushort_uchar    (&xp, nelems, (uchar*)     getbuf);
    else if (gettype == MPI_SHORT)
        return ncmpix_getn_ushort_short    (&xp, nelems, (short*)     getbuf);
    else if (gettype == MPI_UNSIGNED_SHORT)
        return ncmpix_getn_ushort_ushort   (&xp, nelems, (ushort*)    getbuf);
    else if (gettype == MPI_INT)
        return ncmpix_getn_ushort_int      (&xp, nelems, (int*)       getbuf);
    else if (gettype == MPI_UNSIGNED)
        return ncmpix_getn_ushort_uint     (&xp, nelems, (uint*)      getbuf);
    else if (gettype == MPI_LONG)
        return ncmpix_getn_ushort_long     (&xp, nelems, (long*)      getbuf);
    else if (gettype == MPI_FLOAT)
        return ncmpix_getn_ushort_float    (&xp, nelems, (float*)     getbuf);
    else if (gettype == MPI_DOUBLE)
        return ncmpix_getn_ushort_double   (&xp, nelems, (double*)    getbuf);
    else if (gettype == MPI_LONG_LONG_INT)
        return ncmpix_getn_ushort_longlong (&xp, nelems, (longlong*)  getbuf);
    else if (gettype == MPI_UNSIGNED_LONG_LONG)
        return ncmpix_getn_ushort_ulonglong(&xp, nelems, (ulonglong*) getbuf);
    DEBUG_RETURN_ERROR(NC_EBADTYPE)
}

/*----< ncmpii_x_getn_int() >-------------------------------------------------*/
inline int
ncmpii_x_getn_int(const void   *xp,      /* file buffer of type schar */
                 void         *getbuf,  /* get buffer of type gettype */
                 MPI_Offset    nelems,
                 MPI_Datatype  gettype)
{
    if (gettype == MPI_CHAR || /* assume ECHAR has been checked before */
        gettype == MPI_SIGNED_CHAR)
        return ncmpix_getn_int_schar    (&xp, nelems, (schar*)     getbuf);
    else if (gettype == MPI_UNSIGNED_CHAR)
        return ncmpix_getn_int_uchar    (&xp, nelems, (uchar*)     getbuf);
    else if (gettype == MPI_SHORT)
        return ncmpix_getn_int_short    (&xp, nelems, (short*)     getbuf);
    else if (gettype == MPI_UNSIGNED_SHORT)
        return ncmpix_getn_int_ushort   (&xp, nelems, (ushort*)    getbuf);
    else if (gettype == MPI_INT)
        return ncmpix_getn_int_int      (&xp, nelems, (int*)       getbuf);
    else if (gettype == MPI_UNSIGNED)
        return ncmpix_getn_int_uint     (&xp, nelems, (uint*)      getbuf);
    else if (gettype == MPI_LONG)
        return ncmpix_getn_int_long     (&xp, nelems, (long*)      getbuf);
    else if (gettype == MPI_FLOAT)
        return ncmpix_getn_int_float    (&xp, nelems, (float*)     getbuf);
    else if (gettype == MPI_DOUBLE)
        return ncmpix_getn_int_double   (&xp, nelems, (double*)    getbuf);
    else if (gettype == MPI_LONG_LONG_INT)
        return ncmpix_getn_int_longlong (&xp, nelems, (longlong*)  getbuf);
    else if (gettype == MPI_UNSIGNED_LONG_LONG)
        return ncmpix_getn_int_ulonglong(&xp, nelems, (ulonglong*) getbuf);
    DEBUG_RETURN_ERROR(NC_EBADTYPE)
}

/*----< ncmpii_x_getn_uint() >-------------------------------------------------*/
inline int
ncmpii_x_getn_uint(const void   *xp,      /* file buffer of type schar */
                 void         *getbuf,  /* get buffer of type gettype */
                 MPI_Offset    nelems,
                 MPI_Datatype  gettype)
{
    if (gettype == MPI_CHAR || /* assume ECHAR has been checked before */
        gettype == MPI_SIGNED_CHAR)
        return ncmpix_getn_uint_schar    (&xp, nelems, (schar*)     getbuf);
    else if (gettype == MPI_UNSIGNED_CHAR)
        return ncmpix_getn_uint_uchar    (&xp, nelems, (uchar*)     getbuf);
    else if (gettype == MPI_SHORT)
        return ncmpix_getn_uint_short    (&xp, nelems, (short*)     getbuf);
    else if (gettype == MPI_UNSIGNED_SHORT)
        return ncmpix_getn_uint_ushort   (&xp, nelems, (ushort*)    getbuf);
    else if (gettype == MPI_INT)
        return ncmpix_getn_uint_int      (&xp, nelems, (int*)       getbuf);
    else if (gettype == MPI_UNSIGNED)
        return ncmpix_getn_uint_uint     (&xp, nelems, (uint*)      getbuf);
    else if (gettype == MPI_LONG)
        return ncmpix_getn_uint_long     (&xp, nelems, (long*)      getbuf);
    else if (gettype == MPI_FLOAT)
        return ncmpix_getn_uint_float    (&xp, nelems, (float*)     getbuf);
    else if (gettype == MPI_DOUBLE)
        return ncmpix_getn_uint_double   (&xp, nelems, (double*)    getbuf);
    else if (gettype == MPI_LONG_LONG_INT)
        return ncmpix_getn_uint_longlong (&xp, nelems, (longlong*)  getbuf);
    else if (gettype == MPI_UNSIGNED_LONG_LONG)
        return ncmpix_getn_uint_ulonglong(&xp, nelems, (ulonglong*) getbuf);
    DEBUG_RETURN_ERROR(NC_EBADTYPE)
}

/*----< ncmpii_x_getn_float() >-------------------------------------------------*/
inline int
ncmpii_x_getn_float(const void   *xp,      /* file buffer of type schar */
                 void         *getbuf,  /* get buffer of type gettype */
                 MPI_Offset    nelems,
                 MPI_Datatype  gettype)
{
    if (gettype == MPI_CHAR || /* assume ECHAR has been checked before */
        gettype == MPI_SIGNED_CHAR)
        return ncmpix_getn_float_schar    (&xp, nelems, (schar*)     getbuf);
    else if (gettype == MPI_UNSIGNED_CHAR)
        return ncmpix_getn_float_uchar    (&xp, nelems, (uchar*)     getbuf);
    else if (gettype == MPI_SHORT)
        return ncmpix_getn_float_short    (&xp, nelems, (short*)     getbuf);
    else if (gettype == MPI_UNSIGNED_SHORT)
        return ncmpix_getn_float_ushort   (&xp, nelems, (ushort*)    getbuf);
    else if (gettype == MPI_INT)
        return ncmpix_getn_float_int      (&xp, nelems, (int*)       getbuf);
    else if (gettype == MPI_UNSIGNED)
        return ncmpix_getn_float_uint     (&xp, nelems, (uint*)      getbuf);
    else if (gettype == MPI_LONG)
        return ncmpix_getn_float_long     (&xp, nelems, (long*)      getbuf);
    else if (gettype == MPI_FLOAT)
        return ncmpix_getn_float_float    (&xp, nelems, (float*)     getbuf);
    else if (gettype == MPI_DOUBLE)
        return ncmpix_getn_float_double   (&xp, nelems, (double*)    getbuf);
    else if (gettype == MPI_LONG_LONG_INT)
        return ncmpix_getn_float_longlong (&xp, nelems, (longlong*)  getbuf);
    else if (gettype == MPI_UNSIGNED_LONG_LONG)
        return ncmpix_getn_float_ulonglong(&xp, nelems, (ulonglong*) getbuf);
    DEBUG_RETURN_ERROR(NC_EBADTYPE)
}

/*----< ncmpii_x_getn_double() >-------------------------------------------------*/
inline int
ncmpii_x_getn_double(const void   *xp,      /* file buffer of type schar */
                 void         *getbuf,  /* get buffer of type gettype */
                 MPI_Offset    nelems,
                 MPI_Datatype  gettype)
{
    if (gettype == MPI_CHAR || /* assume ECHAR has been checked before */
        gettype == MPI_SIGNED_CHAR)
        return ncmpix_getn_double_schar    (&xp, nelems, (schar*)     getbuf);
    else if (gettype == MPI_UNSIGNED_CHAR)
        return ncmpix_getn_double_uchar    (&xp, nelems, (uchar*)     getbuf);
    else if (gettype == MPI_SHORT)
        return ncmpix_getn_double_short    (&xp, nelems, (short*)     getbuf);
    else if (gettype == MPI_UNSIGNED_SHORT)
        return ncmpix_getn_double_ushort   (&xp, nelems, (ushort*)    getbuf);
    else if (gettype == MPI_INT)
        return ncmpix_getn_double_int      (&xp, nelems, (int*)       getbuf);
    else if (gettype == MPI_UNSIGNED)
        return ncmpix_getn_double_uint     (&xp, nelems, (uint*)      getbuf);
    else if (gettype == MPI_LONG)
        return ncmpix_getn_double_long     (&xp, nelems, (long*)      getbuf);
    else if (gettype == MPI_FLOAT)
        return ncmpix_getn_double_float    (&xp, nelems, (float*)     getbuf);
    else if (gettype == MPI_DOUBLE)
        return ncmpix_getn_double_double   (&xp, nelems, (double*)    getbuf);
    else if (gettype == MPI_LONG_LONG_INT)
        return ncmpix_getn_double_longlong (&xp, nelems, (longlong*)  getbuf);
    else if (gettype == MPI_UNSIGNED_LONG_LONG)
        return ncmpix_getn_double_ulonglong(&xp, nelems, (ulonglong*) getbuf);
    DEBUG_RETURN_ERROR(NC_EBADTYPE)
}

/*----< ncmpii_x_getn_int64() >-------------------------------------------------*/
inline int
ncmpii_x_getn_int64(const void   *xp,      /* file buffer of type schar */
                 void         *getbuf,  /* get buffer of type gettype */
                 MPI_Offset    nelems,
                 MPI_Datatype  gettype)
{
    if (gettype == MPI_CHAR || /* assume ECHAR has been checked before */
        gettype == MPI_SIGNED_CHAR)
        return ncmpix_getn_int64_schar    (&xp, nelems, (schar*)     getbuf);
    else if (gettype == MPI_UNSIGNED_CHAR)
        return ncmpix_getn_int64_uchar    (&xp, nelems, (uchar*)     getbuf);
    else if (gettype == MPI_SHORT)
        return ncmpix_getn_int64_short    (&xp, nelems, (short*)     getbuf);
    else if (gettype == MPI_UNSIGNED_SHORT)
        return ncmpix_getn_int64_ushort   (&xp, nelems, (ushort*)    getbuf);
    else if (gettype == MPI_INT)
        return ncmpix_getn_int64_int      (&xp, nelems, (int*)       getbuf);
    else if (gettype == MPI_UNSIGNED)
        return ncmpix_getn_int64_uint     (&xp, nelems, (uint*)      getbuf);
    else if (gettype == MPI_LONG)
        return ncmpix_getn_int64_long     (&xp, nelems, (long*)      getbuf);
    else if (gettype == MPI_FLOAT)
        return ncmpix_getn_int64_float    (&xp, nelems, (float*)     getbuf);
    else if (gettype == MPI_DOUBLE)
        return ncmpix_getn_int64_double   (&xp, nelems, (double*)    getbuf);
    else if (gettype == MPI_LONG_LONG_INT)
        return ncmpix_getn_int64_longlong (&xp, nelems, (longlong*)  getbuf);
    else if (gettype == MPI_UNSIGNED_LONG_LONG)
        return ncmpix_getn_int64_ulonglong(&xp, nelems, (ulonglong*) getbuf);
    DEBUG_RETURN_ERROR(NC_EBADTYPE)
}

/*----< ncmpii_x_getn_uint64() >-------------------------------------------------*/
inline int
ncmpii_x_getn_uint64(const void   *xp,      /* file buffer of type schar */
                 void         *getbuf,  /* get buffer of type gettype */
                 MPI_Offset    nelems,
                 MPI_Datatype  gettype)
{
    if (gettype == MPI_CHAR || /* assume ECHAR has been checked before */
        gettype == MPI_SIGNED_CHAR)
        return ncmpix_getn_uint64_schar    (&xp, nelems, (schar*)     getbuf);
    else if (gettype == MPI_UNSIGNED_CHAR)
        return ncmpix_getn_uint64_uchar    (&xp, nelems, (uchar*)     getbuf);
    else if (gettype == MPI_SHORT)
        return ncmpix_getn_uint64_short    (&xp, nelems, (short*)     getbuf);
    else if (gettype == MPI_UNSIGNED_SHORT)
        return ncmpix_getn_uint64_ushort   (&xp, nelems, (ushort*)    getbuf);
    else if (gettype == MPI_INT)
        return ncmpix_getn_uint64_int      (&xp, nelems, (int*)       getbuf);
    else if (gettype == MPI_UNSIGNED)
        return ncmpix_getn_uint64_uint     (&xp, nelems, (uint*)      getbuf);
    else if (gettype == MPI_LONG)
        return ncmpix_getn_uint64_long     (&xp, nelems, (long*)      getbuf);
    else if (gettype == MPI_FLOAT)
        return ncmpix_getn_uint64_float    (&xp, nelems, (float*)     getbuf);
    else if (gettype == MPI_DOUBLE)
        return ncmpix_getn_uint64_double   (&xp, nelems, (double*)    getbuf);
    else if (gettype == MPI_LONG_LONG_INT)
        return ncmpix_getn_uint64_longlong (&xp, nelems, (longlong*)  getbuf);
    else if (gettype == MPI_UNSIGNED_LONG_LONG)
        return ncmpix_getn_uint64_ulonglong(&xp, nelems, (ulonglong*) getbuf);
    DEBUG_RETURN_ERROR(NC_EBADTYPE)
}


